
#include "fitting.h"

#include <tjutils/tjtest.h>

#ifdef HAVE_LIBGSL

#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_blas.h>
#include <tjutils/tjnumeric.h>

// data struct to be passed through GSL functions
struct ModelData {
  ModelData(unsigned int size) {n=size; y=new float[size]; sigma=new float[size]; x=new float[size];}
  ~ModelData() {delete[] y; delete[] sigma; delete[] x;}

  ModelFunction* modelfunc;  // model function
  unsigned int n; // number of points
  float* y;     // function values
  float* sigma; // uncertainty of function values
  float* x;     // arguments of function
};


struct GslData4Fit {
  gsl_multifit_fdfsolver *solver;
  gsl_matrix *covar;
};


////////////////////////////////////////////////////////////////

bool ModelFunction::fit(const Array<float,1>& yvals,
         const Array<float,1>& ysigma,
         const Array<float,1>& xvals,
         unsigned int max_iterations, double tolerance) {
  FunctionFit ff(*this,yvals.extent(firstDim),max_iterations,tolerance);
  return ff.fit(yvals,ysigma,xvals);
}


Array<float,1> ModelFunction::get_function(const Array<float,1>& xvals) const {
  int n=xvals.extent(firstDim);
  Array<float,1> result(n);
  for(int i=0; i<n; i++) {
    result(i)=evaluate_f(xvals(i));
  }
  return result;
}

const Array<float,1> ModelFunction::defaultArray;


////////////////////////////////////////////////////////////////

// functions to be used by GSL solver

int FunctionFit_func_f (const gsl_vector * gv, void *params, gsl_vector * f) {

  // get current model function
  ModelData* data=(ModelData*)params;
  ModelFunction* modelfunc=data->modelfunc;
  unsigned int n_fitpars=modelfunc->numof_fitpars();

  // set the fitting parameters in the model function
  for(unsigned int i=0; i<n_fitpars; i++) {
    modelfunc->get_fitpar(i).val=gsl_vector_get(gv, i);
  }

  // calculate vector of values for this set of parameters
  for (unsigned int i = 0; i < data->n; i++) {
    float Yi = modelfunc->evaluate_f(data->x[i]);
    gsl_vector_set (f, i, (data->y[i] - Yi)/data->sigma[i]);
  }

  return GSL_SUCCESS;
}

int FunctionFit_func_df (const gsl_vector * gv, void *params, gsl_matrix * df) {

  // get current model function
  ModelData* data=(ModelData*)params;
  ModelFunction* modelfunc=data->modelfunc;
  unsigned int n_fitpars=modelfunc->numof_fitpars();

  // set the fitting parameters in the model function
  for(unsigned int i=0; i<n_fitpars; i++) {
    modelfunc->get_fitpar(i).val=gsl_vector_get(gv, i);
  }

  // calculate vector of derivative values for this set of parameters
  fvector dYi(n_fitpars);
  for (unsigned int i = 0; i < data->n; i++) {
    dYi = modelfunc->evaluate_df(data->x[i]);
    float s = data->sigma[i];
    for(unsigned int j=0; j<n_fitpars; j++) {
      gsl_matrix_set (df, i, j, -dYi[j]/s);
    }
  }

  return GSL_SUCCESS;
}

int FunctionFit_func_fdf (const gsl_vector * gv, void *params, gsl_vector * f, gsl_matrix * df) {
  FunctionFit_func_f (gv, params, f);
  FunctionFit_func_df (gv, params, df);
  return GSL_SUCCESS;
}

////////////////////////////////////////////////////////////////


FunctionFit::FunctionFit(ModelFunction& model_func, unsigned int nvals, unsigned int max_iterations, double tolerance)
 : func(model_func), max_it(max_iterations), tol(tolerance) {

  data4fit= new ModelData(nvals);

  data4fit->modelfunc=&model_func;

  gsldata=new GslData4Fit;

  // allocate and initialize GSL stuff
  gsldata->covar  = gsl_matrix_alloc (func.numof_fitpars(), func.numof_fitpars());
  gsldata->solver = gsl_multifit_fdfsolver_alloc (gsl_multifit_fdfsolver_lmsder, data4fit->n, func.numof_fitpars());
}

FunctionFit::~FunctionFit() {
  // free GSL stuff
  gsl_multifit_fdfsolver_free(gsldata->solver);
  gsl_matrix_free(gsldata->covar);
  delete data4fit;
  delete gsldata;
}


bool FunctionFit::fit(const Array<float,1>& yvals,
         const Array<float,1>& ysigma,
         const Array<float,1>& xvals) {
  Log<OdinData> odinlog("FunctionFit","fit");

  bool haveSigma;
  bool haveXvals;

  // check input
  if(data4fit->n!=(unsigned)yvals.size() || data4fit->n<=0) {
    ODINLOG(odinlog,errorLog) << "size mismatch in yvals" << STD_endl;
    return false;
  }
  if((unsigned)ysigma.size()==data4fit->n) haveSigma=true;
  else {
    haveSigma=false;
    ODINLOG(odinlog,normalDebug) << "no error bars provided, taking uniform interval for all x" << STD_endl;
  }
  if((unsigned)xvals.size()==data4fit->n) haveXvals=true;
  else {
    haveXvals=false;
    ODINLOG(odinlog,normalDebug) << "no x values provided, taking index as x value" << STD_endl;
  }

  ODINLOG(odinlog,normalDebug) << "n/numof_fitpars=" << data4fit->n << "/" <<  func.numof_fitpars() << STD_endl;

  // copy Blitz arrays to data4fit
  for (unsigned long i=0;i<data4fit->n;i++) {
    data4fit->y[i]=yvals(i);
    if(haveSigma) data4fit->sigma[i]=ysigma(i);
    else data4fit->sigma[i]=0.1;
    if(haveXvals) data4fit->x[i]=xvals(i);
    else data4fit->x[i]=(float)i;
  }
  ODINLOG(odinlog,normalDebug) << "copy arrays done" << STD_endl;


  // initialize GSL fitting function
  gsl_multifit_function_fdf gsl_func;
  gsl_func.f=&FunctionFit_func_f;
  gsl_func.df=&FunctionFit_func_df;
  gsl_func.fdf=&FunctionFit_func_fdf;
  gsl_func.n=data4fit->n;
  gsl_func.p=func.numof_fitpars();
  gsl_func.params=(void *)data4fit;
  ODINLOG(odinlog,normalDebug) << "gsl_func done" << STD_endl;

  // starting values
  double x_init[func.numof_fitpars()];
  for(unsigned int i=0; i<func.numof_fitpars(); i++) {
    x_init[i]=func.get_fitpar(i).val;
  }
  gsl_vector_view initial_guess = gsl_vector_view_array (x_init, func.numof_fitpars());
  ODINLOG(odinlog,normalDebug) << "starting values done" << STD_endl;

  // initialize solver
  gsl_multifit_fdfsolver_set (gsldata->solver, &gsl_func, &initial_guess.vector);
  ODINLOG(odinlog,normalDebug) << "initialize solver done" << STD_endl;


  unsigned int iter = 0;
  int status;
  do {   // main loop
    ODINLOG(odinlog,normalDebug) << "iter= " << iter << STD_endl;
    iter++;
    status = gsl_multifit_fdfsolver_iterate (gsldata->solver);
    ODINLOG(odinlog,normalDebug) << "status(fit)= " << gsl_strerror (status) << STD_endl;
    print_state (iter);
    if (status) break;
    status = gsl_multifit_test_delta (gsldata->solver->dx, gsldata->solver->x, tol, tol);
    ODINLOG(odinlog,normalDebug) << "status(test)= " << gsl_strerror (status) << STD_endl;
    ODINLOG(odinlog,normalDebug) << "iter/max_it= " << iter << "/" << max_it << STD_endl;
  } while (status == GSL_CONTINUE && iter < max_it);

  gsl_multifit_covar (gsldata->solver->J, 0.0, gsldata->covar);

  // copy results
  for(unsigned int i=0;i<func.numof_fitpars();i++) {
    func.get_fitpar(i).val=gsl_vector_get(gsldata->solver->x, i);
    func.get_fitpar(i).err=sqrt(gsl_matrix_get(gsldata->covar,i,i));
  }

  return true;
}



void FunctionFit::print_state (size_t iter) {
#ifdef ODIN_DEBUG
  Log<OdinData> odinlog("FunctionFit","fit");
  ODINLOG(odinlog,normalDebug) << "iter=" << iter << STD_endl;
  for(unsigned int i=0;i<func.numof_fitpars();i++) {
    ODINLOG(odinlog,normalDebug) << "  x" << i << "=" <<  gsl_vector_get (gsldata->solver->x,i) << STD_endl;
  }
  ODINLOG(odinlog,normalDebug) << " |f(x)|=" << fabs(gsl_blas_dnrm2 (gsldata->solver->f)) << STD_endl;
#endif
}

const Array<float,1> FunctionFit::defaultArray;



///////////////////////////////////////////////////////////////////

float ExponentialFunction::evaluate_f(float x) const {
  return A.val * exp (lambda.val * x);
}

fvector ExponentialFunction::evaluate_df(float x) const {
  fvector result(numof_fitpars());
  result[0]= exp (lambda.val * x);
  result[1]= x * A.val * exp (lambda.val * x);
  return result;
}

unsigned int ExponentialFunction::numof_fitpars() const {return 2;}

fitpar& ExponentialFunction::get_fitpar(unsigned int i) {
  if(i==0) return A;
  if(i==1) return lambda;
  return dummy_fitpar;
}

///////////////////////////////////////////////////////////////////


float ExponentialFunctionWithOffset::evaluate_f(float x) const {
  return A.val * exp (lambda.val * x) + C.val;
}

fvector ExponentialFunctionWithOffset::evaluate_df(float x) const {
  fvector result(numof_fitpars());
  result[0]=exp (lambda.val * x);
  result[1]=x * A.val * exp (lambda.val * x);
  result[2]=1.0;
  return result;
}

unsigned int ExponentialFunctionWithOffset::numof_fitpars() const {return 3;}

fitpar& ExponentialFunctionWithOffset::get_fitpar(unsigned int i) {
  if(i==0) return A;
  if(i==1) return lambda;
  if(i==2) return C;
  return dummy_fitpar;
}


///////////////////////////////////////////////////////////////////

float GaussianFunction::evaluate_f(float x) const {
  float arg= (x-x0.val) / fwhm.val;
  return A.val * exp( -2.0 * arg * arg );
}

fvector GaussianFunction::evaluate_df(float x) const {
  fvector result(numof_fitpars());

  float arg= (x-x0.val) / fwhm.val;
  float expterm=exp( -2.0 * arg * arg );

  result[0]= expterm;
  result[1]= 4.0*A.val/(fwhm.val*fwhm.val)*(x-x0.val)*expterm;
  result[2]= 4.0*A.val/(fwhm.val*fwhm.val*fwhm.val)*(x-x0.val)*(x-x0.val)*expterm;
  return result;
}

unsigned int GaussianFunction::numof_fitpars() const {return 3;}

fitpar& GaussianFunction::get_fitpar(unsigned int i) {
  if(i==0) return A;
  if(i==1) return x0;
  if(i==2) return fwhm;
  return dummy_fitpar;
}

///////////////////////////////////////////////////////////////////

float SinusFunction::evaluate_f(float x) const {
  return A.val * sin (m.val * x + c.val);
}

fvector SinusFunction::evaluate_df(float x) const {
  fvector result(numof_fitpars());
  result[0]= sin (m.val * x + c.val);
  result[1]= A.val * x * cos (m.val * x + c.val);
  result[2]= A.val * cos (m.val * x + c.val);
  return result;
}

unsigned int SinusFunction::numof_fitpars() const {return 3;}

fitpar& SinusFunction::get_fitpar(unsigned int i) {
  if(i==0) return A;
  if(i==1) return m;
  if(i==2) return c;
  return dummy_fitpar;
}


///////////////////////////////////////////////////////////////////

bool LinearFunction::fit(const Array<float,1>& yvals,
           const Array<float,1>& ysigma,
           const Array<float,1>& xvals) {
  Log<OdinData> odinlog("LinearFunction","fit");

  m=c=fitpar(); // reset

  int n=yvals.size();
  ODINLOG(odinlog,normalDebug) << "n=" << n << STD_endl;


  if(n<2) {
    ODINLOG(odinlog,errorLog) << "n=" << n << " too small" << STD_endl;
    return false;
  }

  bool haveSigma;
  bool haveXvals;

  if(ysigma.size()==n) haveSigma=true;
  else {
    haveSigma=false;
    ODINLOG(odinlog,normalDebug) << "no error bars provided, taking uniform interval for all x" << STD_endl;
  }
  if(xvals.size()==n) haveXvals=true;
  else {
    haveXvals=false;
    ODINLOG(odinlog,normalDebug) << "no x values provided, taking index as x value" << STD_endl;
  }

  Array<float,1> s(n);
  Array<float,1> x(n);

  if(haveSigma) s=ysigma;
  else s=1.0;

  if(haveXvals) x=xvals;
  else for(int i=0; i<n; i++) x(i)=float(i);

  if(n==2) {
    m.val=secureDivision(yvals(1)-yvals(0), x(1)-x(0));
    c.val=yvals(0)-x(0)*m.val;
    return true;
  }

  Array<float,1> s2(n);
  s2=s*s;

  float S=  sum(1.0/(s2));
  float Sx= sum(x  /(s2));
  float Sxx=sum(x*x/(s2));
  ODINLOG(odinlog,normalDebug) << "S/Sx/Sxx=" << S << "/" << Sx << "/" << Sxx << STD_endl;

  float Sy= sum(yvals/(s2));
  float Sxy=sum(x*yvals/(s2));
  ODINLOG(odinlog,normalDebug) << "Sy/Sxy=" << Sy << "/" << Sxy << STD_endl;

  float Delta=S*Sxx - Sx*Sx;
  ODINLOG(odinlog,normalDebug) << "Delta=" << Delta << STD_endl;

  float beta=secureDivision(S*Sxy-Sx*Sy, Delta);
  float alpha=secureDivision(Sxx*Sy-Sx*Sxy, Delta);
  ODINLOG(odinlog,normalDebug) << "alpha/beta=" << beta << "/" << beta << STD_endl;

  Array<float,1> summand(n);
  summand=(yvals-beta*x-alpha);
  summand*=summand;

  float ystdev=sqrt(secureInv(double(n)-2.0)*sum( summand ));
  ODINLOG(odinlog,normalDebug) << "ystdev=" << ystdev << STD_endl;

  float betastdev=ystdev*sqrt(1.0/( sum(x*x) - secureInv(n)*sum(x)*sum(x) ));
  ODINLOG(odinlog,normalDebug) << "betastdev=" << betastdev << STD_endl;

  float alphastdev=betastdev*sqrt( secureInv(n)*sum(x*x) );
  ODINLOG(odinlog,normalDebug) << "alphastdev=" << alphastdev << STD_endl;

  m.val=beta;
  m.err=betastdev;

  c.val=alpha;
  c.err=alphastdev;

  return true;
}

Array<float,1> LinearFunction::get_function(const Array<float,1>& xvals) const {
  return Array<float,1>(m.val * xvals + c.val);
}

const Array<float,1> LinearFunction::defaultArray;



//////////////////////////////////////////////////////////////
// Unit test

#ifndef NO_UNIT_TEST
class FunctionFitTest : public UnitTest {

 public:
  FunctionFitTest() : UnitTest("FunctionFit") {}

 private:
  bool check() const {
    Log<UnitTest> odinlog(this,"check");

    int numoftestvals=5;

    ExponentialFunctionWithOffset expf_offset;
    FunctionFit expfit_offset(expf_offset,numoftestvals);

    Array<float,1> yvals(numoftestvals);
    Array<float,1> ysigma(numoftestvals);
    Array<float,1> xvals(numoftestvals);

    // some arbitrary values for testing
    xvals(0)=0.1; yvals(0)=12.4; ysigma(0)=0.7;
    xvals(1)=1.2; yvals(1)=8.4;  ysigma(1)=1.2;
    xvals(2)=1.9; yvals(2)=6.1;  ysigma(2)=0.9;
    xvals(3)=3.0; yvals(3)=5.0;  ysigma(3)=0.8;
    xvals(4)=4.1; yvals(4)=5.0;  ysigma(4)=1.0;

    expf_offset.A.val=1.0;  // starting value
    expf_offset.lambda.val=-1.0; // starting value
    expf_offset.C.val=1.0;  // starting value

    if(!expfit_offset.fit(yvals,ysigma,xvals)) {
      ODINLOG(odinlog,errorLog) << "ExponentialFunctionWithOffset fit failed" << STD_endl;
      return false;
    } 

    float A_expected=8.74419;
    float A_err_expected=1.32811;
    float lambda_expected=-0.76985;
    float lambda_err_expected=0.37922;
    float C_expected=4.33682;
    float C_err_expected=1.31881;

    STD_string A_str              =ftos(expf_offset.A.val) +"+-"+ftos(expf_offset.A.err);
    STD_string A_expected_str     =ftos(A_expected    ) +"+-"+ftos(A_err_expected);
    STD_string lambda_str         =ftos(expf_offset.lambda.val) +"+-"+ftos(expf_offset.lambda.err);
    STD_string lambda_expected_str=ftos(lambda_expected)+"+-"+ftos(lambda_err_expected);
    STD_string C_str              =ftos(expf_offset.C.val) +"+-"+ftos(expf_offset.C.err);
    STD_string C_expected_str     =ftos(C_expected)     +"+-"+ftos(C_err_expected);

  /*
    // gives different results depending on the GSL version
    if(A_str     !=A_expected_str)      {ODINLOG(odinlog,errorLog) << "FunctionFit_check: ExponentialFunctionWithOffset failed: A="      << A_str      << ", but expected A="      << A_expected_str      << STD_endl; return false;}
    if(lambda_str!=lambda_expected_str) {ODINLOG(odinlog,errorLog) << "FunctionFit_check: ExponentialFunctionWithOffset failed: lambda=" << lambda_str << ", but expected lambda=" << lambda_expected_str << STD_endl; return false;}
    if(C_str     !=C_expected_str)      {ODINLOG(odinlog,errorLog) << "FunctionFit_check: ExponentialFunctionWithOffset failed: C="      << C_str      << ", but expected C="      << C_expected_str      << STD_endl; return false;}
  */


    /////////////////////////////////////////////////////

    int numoftestvals2=100;

    ExponentialFunction expf;
    FunctionFit expfit(expf,numoftestvals2);

    // arbitrary values for testing
    A_expected=44.5;
    lambda_expected=0.78;

    expf.A.val=A_expected;
    expf.lambda.val=lambda_expected;

    yvals.resize(numoftestvals2);
    ysigma.resize(numoftestvals2);
    xvals.resize(numoftestvals2);

    for(int i=0; i<numoftestvals2; i++) xvals(i)=2.0 * double(i)/double(numoftestvals2);

    // check self integrity of fit
    yvals=expf.get_function(xvals);

    ysigma=1.0;

    if(!expfit.fit(yvals,ysigma,xvals)) {
      ODINLOG(odinlog,errorLog) << "ExponentialFunction fit failed" << STD_endl;
      return false;
    }

    A_str              =ftos(expf.A.val);
    A_expected_str     =ftos(A_expected);
    lambda_str         =ftos(expf.lambda.val);
    lambda_expected_str=ftos(lambda_expected);

    if(A_str     !=A_expected_str)      {ODINLOG(odinlog,errorLog) << "ExponentialFunction failed: A="      << A_str      << ", but expected A="      << A_expected_str      << STD_endl; return false;}
    if(lambda_str!=lambda_expected_str) {ODINLOG(odinlog,errorLog) << "ExponentialFunction failed: lambda=" << lambda_str << ", but expected lambda=" << lambda_expected_str << STD_endl; return false;}

    /////////////////////////////////////////////////////

    LinearFunction linf;

    yvals.resize(numoftestvals);
    ysigma.resize(numoftestvals);
    xvals.resize(numoftestvals);

    // some arbitrary values for testing
    xvals(0)=0.1; yvals(0)=2.4;  ysigma(0)=0.7;
    xvals(1)=1.2; yvals(1)=3.4;  ysigma(1)=1.2;
    xvals(2)=1.9; yvals(2)=4.1;  ysigma(2)=0.9;
    xvals(3)=3.0; yvals(3)=5.0;  ysigma(3)=0.8;
    xvals(4)=4.1; yvals(4)=6.0;  ysigma(4)=1.0;

    if(!linf.fit(yvals,ysigma,xvals)) {
      ODINLOG(odinlog,errorLog) << "LinearFunction fit failed" << STD_endl;
      return false;
    }

    float m_expected=0.89863;
    float m_err_expected=0.01351;
    float c_expected=2.32634;
    float c_err_expected=0.03357;

    STD_string m_str              =ftos(linf.m.val) +"+-"+ftos(linf.m.err);
    STD_string m_expected_str     =ftos(m_expected) +"+-"+ftos(m_err_expected);
    STD_string c_str              =ftos(linf.c.val) +"+-"+ftos(linf.c.err);
    STD_string c_expected_str     =ftos(c_expected) +"+-"+ftos(c_err_expected);

    if(m_str     !=m_expected_str)      {ODINLOG(odinlog,errorLog) << "LinearFunction failed: m="      << m_str      << ", but expected m="      << m_expected_str      << STD_endl; return false;}
    if(c_str     !=c_expected_str)      {ODINLOG(odinlog,errorLog) << "LinearFunction failed: c="      << c_str      << ", but expected c="      << c_expected_str      << STD_endl; return false;}

    /////////////////////////////////////////////////////

    const int polydegree=4;
    PolynomialFunction<polydegree> polyf;

    float a_expected[polydegree];
    polyf.a[0].val=a_expected[0]=-4.0;
    polyf.a[1].val=a_expected[1]=0.5;
    polyf.a[2].val=a_expected[2]=0.0;
    polyf.a[3].val=a_expected[3]=-1.0;

    int numoftestvals3=100;
    Array<float,1> xvals3(numoftestvals3);
    Array<float,1> ysigma3(numoftestvals3);
    for(int i=0; i<numoftestvals3; i++) {
      xvals3(i)=float(i)/float(numoftestvals3)-0.5;
      ysigma3(i)=fabs(sin(i+0.5)); // jittered error
    }

    Array<float,1> yvals3(polyf.get_function(xvals3));

//    Data<float,1>(yvals3).autowrite("yvals3.asc");
//    Data<float,1>(ysigma3).autowrite("ysigma3.asc");

    if(!polyf.fit(yvals3,ysigma3,xvals3)) {
      ODINLOG(odinlog,errorLog) << "PolynomialFunction fit failed" << STD_endl;
      return false;
    }

    for(int i=0; i<polydegree; i++) {
      float fitval=polyf.a[i].val;
      float expval=a_expected[i];
      if((fitval-expval)>1.0e-3) {
        ODINLOG(odinlog,errorLog) << "PolynomialFunction failed: a[" << i << "]=" << fitval << ", but expected " << expval << STD_endl;
        return false;
      }
    }


    /////////////////////////////////////////////////////

    Array<float,2> values(3,3);
    values=10.0;
    values(1,1)=0.0;

    Array<float,2> reliability(3,3);
    reliability=1.0;
    reliability(1,1)=0.0;

    Data<float,2> pfresult(3,3);

    pfresult=polyniomial_fit(values,reliability,1,2.0);
    if(fabs(pfresult(1,1)-10.0)>1.0e-3) {
      ODINLOG(odinlog,errorLog) << "values=" << values << STD_endl;
      ODINLOG(odinlog,errorLog) << "reliability=" << reliability << STD_endl;
      ODINLOG(odinlog,errorLog) << "pfresult=" << pfresult << STD_endl;
      ODINLOG(odinlog,errorLog) << "polyniomial_fit failed" << STD_endl;
      return false;
    }


    int testsize=20;
    Data<float,2> testarray(testsize,testsize);
    Data<float,2> testrelia(testsize,testsize);
    TinyVector<int,2> index;
    for(int i=0; i<testarray.numElements(); i++) {
      index=testarray.create_index(i);
      float radius=norm((index(0)-testsize/2),(index(1)-testsize/2));
      radius/=(0.5*float(testsize));
      if(radius<(2.0/3.0)) {
        testarray(index)=radius*radius;
        testrelia(index)=1.0;
      } else {
        testarray(index)=0.0;
        testrelia(index)=0.0;
      }
    }

    pfresult.resize(testsize,testsize);
    pfresult=polyniomial_fit(testarray,testrelia,2,5.0);
    pfresult=pfresult*testrelia;

    float diff=sum(fabs(testarray-pfresult));
    if(diff>1.0e-3) {
      ODINLOG(odinlog,errorLog) << "polyniomial_fit failed, diff=" << diff << STD_endl;
      return false;
    }

    return true;
  }

};

void alloc_FunctionFitTest() {new FunctionFitTest();} // create test instance
#endif



//////////////////////////////////////////////////////////////


#else
#error "GNU Scientific library is missing!"
#endif



